from unittest import TestCase

from simsim import parse


TEST_CODE_BASIC = """
a = 1
"""

TEST_CODE_FUNC = """
n = 'test'

def func_one(x):
    return x*2
"""

TEST_CODE_CLASS = """
class Foo(object):
    def bar(self):
        return True
"""

TEST_CODE_MODULE_DOCSTRING = '""" module docstring """\nx = 10\n'

TEST_CODE_FUNC_DOCSTRING = \
    'def foo(bar):\n' \
    '    """ func docstring """\n' \
    '    return bar\n'

TEST_CODE_CLASS_DOCSTRING = \
    'class Foo(object):\n' \
    '    """ class docstring """\n' \
    '    def bar(self):\n' \
    '        pass\n'


class TestParser(TestCase):
    def test_basic(self):
        expected = {
            '__main__': ['Assign', 'Name', 'Store', 'Num']
        }
        self.assertDictEqual(parse(TEST_CODE_BASIC).scope_nodes, expected)

    def test_func(self):
        expected = {
            '__main__': ['Assign', 'Name', 'Store', 'Str', 'FunctionDef'],
            '__main__.func_one': ['FunctionDef', 'arguments', 'arg',
                                  'Return', 'BinOp', 'Name',
                                  'Load', 'Mult', 'Num'],
        }
        self.assertDictEqual(parse(TEST_CODE_FUNC).scope_nodes, expected)

    def test_class(self):
        expected = {
            '__main__': ['ClassDef'],
            '__main__.Foo': ['ClassDef', 'Name', 'Load', 'FunctionDef'],
            '__main__.Foo.bar': ['FunctionDef', 'arguments', 'arg',
                                 'Return', 'NameConstant'],
        }
        self.assertDictEqual(parse(TEST_CODE_CLASS).scope_nodes, expected)

    def test_docstring(self):
        nodes = parse(TEST_CODE_MODULE_DOCSTRING).scope_nodes
        expected = {
            '__main__': ['Assign', 'Name', 'Store', 'Num']
        }
        self.assertDictEqual(nodes, expected)

        nodes = parse(TEST_CODE_FUNC_DOCSTRING).scope_nodes
        expected = {
            '__main__': ['FunctionDef'],
            '__main__.foo': ['FunctionDef', 'arguments', 'arg',
                             'Return', 'Name', 'Load']
        }
        self.assertDictEqual(nodes, expected)

        nodes = parse(TEST_CODE_CLASS_DOCSTRING).scope_nodes
        expected = {
            '__main__': ['ClassDef'],
            '__main__.Foo': ['ClassDef', 'Name', 'Load', 'FunctionDef'],
            '__main__.Foo.bar': ['FunctionDef', 'arguments', 'arg', 'Pass'],
        }
        self.assertDictEqual(nodes, expected)



from unittest import TestCase
from simsim.cmd import compare


TEST_FUNC_A = """
def func_one(x):
    return x*2
"""

TEST_FUNC_B = """
def func_two(x):
    return x*2
"""


TEST_FUNC_REORDER_A = """
def func_one(x):
    return x*3

def func_two(x, y):
    return bool(x) and bool(y)
"""


TEST_FUNC_REORDER_B = """
def func_two(x, y):
    return bool(x) and bool(y)

def func_one(x):
    return x*3
"""

TEST_CODE_3_A = """
a = 1
b = 3

print(a + b)
"""

TEST_CODE_3_B = """
a = 1 + 0
b = 3

print(a + b)
"""


class TestCompare(TestCase):
    @classmethod
    def assert_compare(cls, a, b, ratio):
        assert compare(a, b)['ratio'] == ratio

    def test_vars(self):
        self.assert_compare('a = 1', 'b = 1', 1)

    def test_funcs(self):
        self.assert_compare(TEST_FUNC_A, TEST_FUNC_B, 1)
        self.assert_compare(TEST_FUNC_REORDER_A, TEST_FUNC_REORDER_B, 1)

    def test_less_one(self):
        assert compare(TEST_CODE_3_A, TEST_CODE_3_B)['ratio'] < 1



import operator
from argparse import ArgumentParser

from simsim import compare


def main():
    def get_file(value):
        return open(value, 'r')

    parser = ArgumentParser()
    parser.add_argument('file', type=get_file, nargs=2)

    args = parser.parse_args()
    file_a, file_b = [f for f in args.file]
    result = compare(file_a.read(), file_b.read())
    print('Similarity: {:.2f}'.format(result.get('ratio', 0)))

    for n, s in result.get('scopes', {}).items():
        label = '\n{}:{}'.format(file_a.name, n)
        print(label)
        print('-'*(len(label)-1))
        for func, sim in sorted(s, key=operator.itemgetter(1), reverse=True):
            print('{}:{}: {:.2f}'.format(file_b.name, func, sim))


if __name__ == '__main__':
    main()



from collections import defaultdict, Sequence
from ast import NodeVisitor, Str, Expr, Module


# noinspection PyPep8Naming
class Checker(NodeVisitor):

    def __init__(self):
        super().__init__()
        self.scopes = []
        self.scope_nodes = defaultdict(list)

    @staticmethod
    def mark_docstring_sub_nodes(node):
        def _mark_docstring_nodes(body):
            if body and isinstance(body, Sequence):
                for n in body:
                    if isinstance(n, Expr) and isinstance(n.value, Str):
                        n.is_docstring = True

        node_body = getattr(node, 'body', None)
        _mark_docstring_nodes(node_body)
        node_orelse = getattr(node, 'orelse', None)
        _mark_docstring_nodes(node_orelse)

    @staticmethod
    def is_docstring(node):
        return getattr(node, 'is_docstring', False)

    def append_to_scope(self, node):
        if self.scopes:
            stack_name = '.'.join(self.scopes)
            self.scope_nodes[stack_name].append(node.__class__.__name__)

    def process_scope(self, node, name=None):
        self.scopes.append(name or node.name)
        self.generic_visit(node)
        self.scopes.pop()

    def generic_visit(self, node):
        self.mark_docstring_sub_nodes(node)

        if self.is_docstring(node):
            return

        if not isinstance(node, Module):
            self.append_to_scope(node)
        return super().generic_visit(node)

    def visit_AsyncFunctionDef(self, node):
        self.append_to_scope(node)
        self.process_scope(node)
        return node

    def visit_FunctionDef(self, node):
        self.append_to_scope(node)
        self.process_scope(node)
        return node

    def visit_ClassDef(self, node):
        self.append_to_scope(node)
        self.process_scope(node)
        return node

    def visit_Module(self, node):
        self.process_scope(node, '__main__')
        return node



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="simsim",
    version="0.0.1",
    author="Dmitry Gerasimenko",
    author_email="kiddima@gmail.com",
    description="Simple python code similarity checker",
    keywords='plagiarism, plage, plagiary, similarity, ast',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/kidig/simsim",
    packages=setuptools.find_packages(),
    zip_safe=False,
    entry_points={
        'console_scripts': ['simsim = simsim.cmd:main'],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)


import ast
from collections import defaultdict
from difflib import SequenceMatcher

from .checker import Checker

__all__ = ['parse', 'compare', 'Checker']

__version__ = '0.0.1'


def parse(source: str) -> 'Checker':
    chk = Checker()
    tree = ast.parse(source)
    # print(ast.dump(tree))
    chk.visit(tree)
    return chk


def compare(a: str, b: str):
    file_a, file_b = map(parse, [a, b])
    points = []
    items = defaultdict(list)

    for ka, va in file_a.scope_nodes.items():
        max_sim = 0
        for kb, vb in file_b.scope_nodes.items():
            sim = SequenceMatcher(None, va, vb).ratio()
            items[ka].append((kb, sim))
            if sim > max_sim:
                max_sim = sim
        points.append(max_sim)

    return {
        'ratio': sum(points) / len(points),
        'scopes': items,
    }
